library(plyr)
library(xtable)
library(seasonal)
library(ggfortify)
library(dplyr)
library(data.table)
library(forecast)
library(readr)
library(stringr)
library(zoo)
library(pracma)

counter_nan <- function(x)
{
return(length(which(is.na(x))))
}

# Computing Aggregate radiance for Bombay 

removeNA <- function(ts)
{
    return(ifelse(ts=="NaN", 0, ts))
}


## This function turns any values below 0 to NA.
threshold_apply <- function(ts)
{
    return(ifelse(ts<=0, NA, ts))
}


## This function removes outliers
OutlierRem <- function(ts) 
{
    if (counter_nan(ts) > 30)
    {
        return(ts)
    }else
    {
        return(tsclean(ts, replace.missing=FALSE)) 
    }
    
}


## This function interpolates the missing data in a time-series.
interpolate <- function(ts) 
{
    if(counter_nan(ts) > 50)
    {
        return(rep(0, length(ts)))
    }else
    {
        return(na.interp(ts))
    }
    
}

## This function applies a given function i.e. func to all the timeseries in a 3d matrix.
# Example - In a matrix of 8000 x 7100 x 93, long_apply() function will apply the given function on 8000 x 7100 pixels
# each of whom have a time-series of 93. 

long_apply <- function(data_cube, func)
{
    tmp <- array(dim=dim(data_cube))
    for (i in 1:dim(data_cube)[1])
    {
        for (j in 1:dim(data_cube)[2])
        {
            tmp[i,j,] <- func(data_cube[i,j,])
        }
    }
    return(tmp)
}

## This function applies a given function i.e. func to all the 2d matrices in a 3d matrix.
# Example - In a matrix of 8000 x 7100 x 93, cross_apply() function will apply the given function on each matrix of (8000 x 7100)

cross_apply <- function(data_cube, func)
{
    tmp <- array(dim=dim(data_cube)[3])
    for(i in 1:dim(data_cube)[3])
    {
        tmp[i] <- func(data_cube[,,i])
    }
    return(tmp)
}

seasonal_adjustment <- function(nsa, foldername="test", outlier=NULL)
{
    nsa.ts <- as.ts(nsa)
    runSA <- seas(nsa.ts,
                  regression.variables = c("const"),
                  regression.aictest = NULL,
                  outlier = outlier,
                  transform.function = "log",
                  x11 = "", out=FALSE)
    sa.ts <- final(runSA)
    sa.zoo <- zoo(sa.ts, order.by = as.yearmon(as.Date(paste("01", as.yearmon(index(nsa)), sep = " "), "%d %b %Y")))
    summsa <- summary(runSA)
    seasfactor <- round(series(runSA, "d10") * 100, 2)
    seasfactor <- fortify(seasfactor)
    seasfactor$Year <- substr(seasfactor$Index, 1, 4)
    seasfactor$Month <- as.factor(substr(as.yearmon(seasfactor$Index), 1, 3))
    seasfactor <- reshape2::dcast(seasfactor, Year ~ Month, value.var = "Data", fill = NA)
    seasfactor <- seasfactor[, c(1, match(month.abb, colnames(seasfactor)))]
    ## Find range
    seasfactor$Range <- NA
    for(i in 1:nrow(seasfactor))
    {
        seasfactor$Range[i] <- max(as.numeric(seasfactor[i, 2:ncol(seasfactor)]), na.rm = T) - min(as.numeric(seasfactor[i, 2:ncol(seasfactor)]), na.rm = T)
    }
    return(list(sa.series = sa.zoo,
                seasonal.factor.range = seasfactor))
}

monthly_seasonality <- function(timeseries_data,  outlier=NULL)
{
    range        <- seasonal_adjustment(timeseries_data,"test",outlier=outlier)$seasonal.factor.range
    avg_range    <- colMeans(range[,-1], na.rm=TRUE) %>% as.data.frame()
    avg_range    <- setDT(avg_range, keep.rownames = TRUE)[]
    colnames(avg_range) <- c("month", "average.range")
    return(avg_range)
}


monthlySeasonality_df <- function(hr_raw_radiance, hr_conventional_radiance, hr_bc_radiance, wb_hr_radiance, hr_income)
{
        df                      <- data.frame(month = monthly_seasonality(ts(hr_raw_radiance,start = c(2014, 01),frequency=12))$month,
                                        raw = monthly_seasonality(ts(hr_raw_radiance,start = c(2014, 01),frequency=12))$average.range, stringsAsFactors=FALSE)
        df$income               <- monthly_seasonality(ts(hr_income,start = c(2014,01),frequency=12))$average.range
        df$conventional         <- monthly_seasonality(ts(hr_conventional_radiance,start = c(2014, 01),frequency=12))$average.range
        df$bc                   <- monthly_seasonality(ts(hr_bc_radiance,start = c(2014, 01),frequency=12))$average.range
        df$cleanedWb            <- monthly_seasonality(ts(wb_hr_radiance,start = c(2014, 01),frequency=12))$average.range
        df                      <- df[-nrow(df),]  ## Removing "Range"
        return(df)
}



## Bias correction

bias_correction <- function(radiances,CFObs,cutoff=1)
{
    if(cor_test(radiances, CFObs) > 0.05){
        return(radiances) # Don't perform correction if there is no correlation between the radiance and cfobs. 
    }
    cf           <- CFObs/max(CFObs)
    ly1          <- log(radiances)            
    date         <- seq(as.yearmon("2014-01"), by=1/12, length.out=length(ly1))
    time         <- date - as.yearmon("2014-01")
    m.timetrend  <- lm(ly1 ~ time,na.action=na.exclude)
    ly2          <- residuals(m.timetrend)           # ly2 is the residuals after removing the time trend
    m      <- loess(ly2 ~ cf,na.action = "na.exclude")
    topval              <- max(na.omit(predict(m)))
    corrected           <- ly2 + topval - predict(m)
    fixed.1             <- ly2
    fixed.1             <- corrected 
    fixed.2             <- fixed.1 + (coef(m.timetrend)[2] * time) + coef(m.timetrend)[1]
    fixed.3             <- ly1 
    tocorrect           <- fixed.2 < max(na.omit(ly1))
    tocorrect[is.na(tocorrect)]<-FALSE
    fixed.3[tocorrect]  <- fixed.2[tocorrect]
    return(exp(fixed.3))
}

bias_correction_ts <- function(radiance, cfobs, mask)
{
    rad_cleaned <- radiance
    for(i in 1:dim(radiance)[1])
    {
        for(j in 1:dim(radiance)[2])
        {
            if (counter_nan(radiance[i,j,]) > 50)
            {
                rad_cleaned[i,j,] <- radiance[i,j,]
            }else if(mask[i,j]==0)
            {
                rad_cleaned[i,j,] <- radiance[i,j,]
            }else
            {

                rad_cleaned[i,j,] <- bias_correction(radiance[i,j,], cfobs[i,j,])
            }

        }
    }
    return(rad_cleaned)
}

detrend_ts <- function(x){
    date         <- seq(as.yearmon("2014-01"), by=1/12, length.out=length(x))
    time         <- date - as.yearmon("2014-01")
    m.timetrend  <- lm(x ~ time,na.action=na.exclude)
    
    return(residuals(m.timetrend)) 
}
    
cor_test <- function(x,y)
{
    a           <- data.frame(Cleaned1 = x, CF.obs= y)
    colnames(a) <- c("Cleaned1","CF.obs")
    a$date       <- seq(as.yearmon("2012-04"), by=1/12, length.out=nrow(a))
    a$time       <- a$date - as.yearmon("2012-04")
    a$ly1        <- (a$Cleaned1)
    m            <- lm(ly1 ~ time, data=a,na.action=na.exclude)
    a$ly2        <- residuals(m) #removing time trend

    ot           = cor.test(a$CF.obs,a$ly2, alternative = "greater",method="spearman",exact=FALSE,na.action=na.omit) # one tail test
    return (ot$p.value)
}

apply_mask <- function(ts, mask)
{
    tmp <- ts
    for (i in 1:dim(ts)[3])
    {
        tmp[,,i] <- ts[,,i]*mask
    }
    return(tmp)
}


mark_na <- function(ts, clouds){
for (i in 1:dim(ts)[1]){
    for (j in 1:dim(ts)[2]){
        for (k in 1:dim(ts)[3])
            if(clouds[i, j, k] == 0){
                ts[i, j, k] = NA
            }

    }
}
return(ts)
}
noise_mask_generator <- function(ts, clouds, mask=matrix(1, dim(radiance_Bombay)[1], dim(radiance_Bombay)[2]), threshold = 0.4)
{
    weighted_radiance        <- matrix(0, dim(ts)[1], dim(ts)[2])
    noise_mask <- matrix(0, dim(ts)[1], dim(ts)[2])
    for (i in 1:dim(ts)[1])
    {
        for (j in 1:dim(ts)[2])
        {
            if (mask[i,j]== 0)
            {
                weighted_radiance[i,j]=0
            }else if (counter_nan(ts[i,j,] > 50))
            {
                weighted_radiance[i,j]= 0
            }else
            {
                weighted_radiance[i,j] = weighted.mean(ts[i,j,], clouds[i,j,], na.action = na.exclude)
            }
            
        }
    }
    for (i in 1:dim(weighted_radiance)[1])
    {
        for (j in 1:dim(weighted_radiance)[2])
        {   
            if(weighted_radiance[i,j] < threshold){
                noise_mask[i,j] = 1 
            }
        }
    }
    return(noise_mask)
}

outlier_mask_generator <- function(ts, mask)
{
    stds        <- matrix(0, dim(ts)[1], dim(ts)[2])
    outlierMask <- matrix(0, dim(ts)[1], dim(ts)[2])
    for (i in 1:dim(ts)[1])
    {
        for (j in 1:dim(ts)[2])
        {
            if (mask[i,j]==0)
            {
                stds[i,j]= 0 
            }
            else if (counter_nan(ts[i,j,]) > 50)
            {
                stds[i,j]= 0 
            }
            
            else
            {
                stds[i,j] = sd(detrend_ts(ts[i,j,]), na.rm=TRUE)
            }
            
        }
    }

    std_mask <- function(x)
    {
        if(x < 11.9)
        {
            return(1)
        }else
        {
            return(0)
        }
    }
    for (i in 1:dim(stds)[1])
    {
        for (j in 1:dim(stds)[2])
        {
            outlierMask[i,j] <- std_mask(stds[i,j])
        }
    }
    outlierMask     <- outlierMask*mask
    return(outlierMask)
}


### Plots

library(zoo)
library(dplyr)
library(latex2exp)
options(scipen=999)
# color and line type coding for all the plots.

col_rawRad              <- "green4"
lty_rawRad              <- 2
col_conventionalRad     <- "red"
lty_conventionalRad     <- 1
col_bcRad               <- "blue"
lty_bcRad               <- 3
col_worldbankRad        <- "purple3"
lty_worldbankRad        <- 2
col_income              <- "black"
lty_income              <- 1
col_seasonallyAdj       <- "green"
lty_seasonallyAdj       <- 1


scale = 1.65
cexlab      = 1.3*scale
cexlegend   = 1.3*scale
cexaxis     = 1.3*scale
linewidth   = 3
imageWidth  = 800
margins     = c(1.0,1.2,.4,.4)

rawLeg = "Raw"
conLeg = "Conventional"
bcLeg  = "New"
incLeg = "Income"
wbLeg  = "World Bank"

radUnits = TeX('Radiance ($nW cm^{-2}$ $sr^{-1})$')
radUnitsLog = TeX('Detrended log radiance ($nW cm^{-2}$ $sr^{-1})$')

#install.packages("extrafont")
library(extrafont)
#font_install('fontcm')
TexFont = "LM Roman 10"


## Function to generate radiance plot for 2 series.
radiance_plot_2series <- function(radiance, radiance2, pngFilename, title, legendTitle, startMonth="2014-01", lty_type=c(1,2), ymin=min(c(radiance, radiance2)), ymax=max(c(radiance, radiance2))*1.3, color=c("black","blue"))
{
    png(pngFilename, width=imageWidth)
    par(mfrow=c(1,1))
    par(mai=margins)
    par(family = TexFont)
    yrange  <- range(c(ymin, ymax))
    months  <- seq(as.yearmon(startMonth), length.out=length(radiance), by=1/12)
    plot(zoo(radiance, order.by = months), type="l", lwd=linewidth, ylim = yrange, xlab="", cex.axis=cexaxis,
                ylab=radUnits, main = title, cex.lab=cexlab, cex.main=2, col=color[1], lty=lty_type[1])
    lines(zoo(radiance2, order.by = months), type="l", lwd=linewidth, col=color[2], lty=lty_type[2])
    legend(x="topright", bty="n", lwd=linewidth, col=color, legend=legendTitle, cex=cexlegend, lty=lty_type)
    abline(h=0)
    grid(lwd=linewidth)
    dev.off()

}
## Function to generate radiance and income plot.
radiance_income_plot <- function(radiance, income, legendTitle, pngFilename, startMonth="2014-01", color=c("black","blue"), lty_type=c(1,1))
{
    png(pngFilename, width=imageWidth)
    par(mai=c(.6,1.2,.4,1.0))
    par(family = TexFont)
    monthsCp          <- seq(as.yearmon(startMonth), length.out=length(income), by=1/12)
    plot(zoo(radiance, order.by = monthsCp), type="l", lwd=linewidth, xlab="", cex.axis=cexaxis, ylim=c(0, max(radiance)*1.35),
                 ylab=radUnits, main = "", cex.lab=cexlab, lty=lty_type[1], col=color[1])

    par(new = TRUE)                             # Add new plot
    plot(zoo(income, order.by = monthsCp), type="l", axes=FALSE, lwd=6, lty=lty_type[2],
                 col=color[2], xlab="", cex.axis=cexaxis, ylab="", ylim=c(0, max(income)*1.35), cex.lab=cexlab)

    axis(side = 4, at = pretty(range(income)), cex.axis=cexaxis)      # Add second axis
    mtext("INR (in million)", side = 4, line = 3, cex=cexlab)
    legend(x="topright", bty="n", lwd=c(3,6), col=color, legend=legendTitle, cex=cexlegend, lty=lty_type)
    abline(h=0)
    grid(lwd=linewidth)
    dev.off()
}

## Function to generate a plot of radiance and cloud-free observation series.
radiance_cfobs_plot <- function(radiance, cfobs, pngFilename, startMonth="2014-01", vline=FALSE, color=c("blue", "black"), lty_type)
{
    png(pngFilename, width=imageWidth)
    par(mfrow=c(2,1))
    par(mai=c(0,1.2,.4,.4)*1.2)
    par(family = TexFont)
    
    y_max = max(radiance)*1.2
    y_min = min(radiance)*1.1
    yrange  <- range(c(min(radiance), radiance), na.rm=TRUE)
    yrange  <- yrange
    months  <- seq(as.yearmon(startMonth), length.out=length(radiance), by=1/12)

    plot(zoo(radiance, order.by=months), xlab="", lwd=linewidth, cex.lab=1.9, col=color[1], lty=lty_type[1], ylim=yrange, xaxt="n", ylab="", cex.axis=cexaxis)

    abline(h=0)
    if(vline==TRUE)
    {
        # legend(x="topright", bty="n", lwd=linewidth, lty=2, legend="July", col="brown", cex=cexlegend)
        abline(v=months[grepl("Jul", months)], col="brown", lwd=2, lty=2)
    }
    par(new=T)
    plot(zoo(radiance, order.by=months), type="p", cex=(cfobs==0)+0, pch=19, cex.lab=1.9,
            col=color[2], xlab="",ylab="", ylim=yrange, xaxt="n", yaxt="n")
    ## Paint the number of images, and flag off the months with 0 cloud free images
    mtext(text = TeX('Radiance'),
      side = 2,#side 1 = bottom
      line = 5,
      cex =1.9)
    mtext(text = TeX('($nW cm^{-2}$ $sr^{-1})$'),
      side = 2,#side 1 = bottom
      line = 2.5,
      cex =1.9)
    par(mai=c(1,1.2,0,.4)*1.2)
    plot(zoo(cfobs, order.by=months), xlab="", type="h", cex.lab=cexlab, ylab="Cloud-free \n observations", lwd=5, xaxt="n", yaxt="n")
    abline(h=0)
    if(vline==TRUE)
    {
        abline(v=months[grepl("Jul", months)], col="brown", lwd=2, lty=2)
    }
    par(new=T)
    plot(zoo(cfobs, order.by=months), type="p", cex=(cfobs==0)+0, pch=19,cex.lab=1.3, col="black", ylab="", xlab="", cex.axis=cexaxis)
    dev.off()
}

# Function to generate 1 lineplot of monthly seasonality of an area.
monthlySeas_lineplot  <- function(df, legendTitle, pngFilename, color=c("#009999","#0000FF"), lty_type)
{
    png(pngFilename, width = 800)
    par(mai=margins)
    par(family = TexFont)
    df$month <- as.Date(c("2020-01-01", "2020-02-01", "2020-03-01", "2020-04-01", "2020-05-01", "2020-06-01", "2020-07-01", "2020-08-01", "2020-09-01", "2020-10-01", "2020-11-01", "2020-12-01")) 
    plot(df$month, (df$income-100), type="b", col=color[1], ylim=c(-30,30), lty=lty_type[1], ylab="% deviation from ann. mean", xlab="", cex.axis=cexaxis,cex.lab=cexlab, lwd=6)
    lines(df$month, (df$conventional-100), type="l", col=color[2], lwd=linewidth, lty=lty_type[2])
    legend(x="topright", bty="n", lwd=linewidth, col=color, legend=legendTitle, cex=cexlegend, lty=lty_type,pch=c(21,NA))
    grid(lwd=linewidth)
    dev.off()
}



## Function to generate 3 lineplots(income, conventional cleaning, bias corrected and World Bank radiance) of monthly seasonality of an HR.
monthlySeas_lineplot_4series  <- function(df, legendTitle, pngFilename, title="", color=c("#009999", "#0000FF", "red", "#339900"), lty_type, ymin=-30, ymax=55, incomeLegend_yloc=41)
{
    y_max = (max(c(df$conventional-100,df$bc-100, df$cleanedWb -100, df$income -100)))*1.3
    y_min = (min(c(df$conventional-100,df$bc-100)))*1.2
    linewidthtmp = 4
    png(pngFilename, width = 800)
    par(family = TexFont)
    par(mai=margins)
    df$month <- as.Date(c("2020-01-01", "2020-02-01", "2020-03-01", "2020-04-01", "2020-05-01", "2020-06-01", "2020-07-01", "2020-08-01", "2020-09-01", "2020-10-01", "2020-11-01", "2020-12-01")) 
    plot(df$month, (df$conventional-100), type="l", col=color[4], main=title, cex.main=2,lwd=linewidthtmp, lty=lty_type[4], ylim=c(y_min, y_max), ylab="% deviation from ann. mean", xlab="", cex.axis=cexaxis,cex.lab=cexlab)
    lines(df$month, (df$bc-100), type="l", col=color[2], lwd=6, lty=lty_type[2])
    lines(df$month, (df$cleanedWb-100), type="l", col=color[3], lwd=linewidthtmp, lty=lty_type[3])
    lines(df$month, (df$income-100), type="b", col=color[1], lwd=6, lty=lty_type[1])
    legend(x="bottomleft", bty="n", lwd=c(linewidthtmp,6,linewidthtmp,6), col=color, legend=legendTitle, cex=cexlegend, lty=lty_type,pch=c(21,NA,NA,NA))
    grid(lwd=linewidth)
    dev.off()
}

