source("Commons.R")
library(rhdf5)

## Reading JLD files

conventionalNTL     <- h5read("../RESULTS/conventional_datacube_Bombay.jld","data")
biasCorrectedNTL    <- h5read("../RESULTS/new_datacube_Bombay.jld","data")
radiance_Bombay     <- h5read("../DATA/raw_radiance_datacube_Bombay.jld", "data") 
clouds_Bombay       <-  h5read("../DATA/cfobs_datacube_Bombay.jld", "data")  #number of cloud-free observations for each pixel 

## Reading income and WorldBank radiance and sample pixel data
Bombay              <- read_csv("../DATA/other_timeseries.csv")


## Aggregating values

rawAgg              <- cross_apply(removeNA(radiance_Bombay), sum)
conventionalAgg     <- cross_apply(conventionalNTL,sum)
biasCorrectedAgg    <- cross_apply(biasCorrectedNTL,sum)
cloudsAgg           <- cross_apply(clouds_Bombay, sum)
incomeAgg           <- Bombay$income/1e6
worldbankAgg        <- Bombay$WorldBank_radiance


## Calculating values for the example pixel in Bombay
pixel  <- data.frame(radiance = Bombay$samplePixel_radiance, n.images = Bombay$samplePixel_clouds)

## Step 2 - Replace 0 with NA

# Step 3 - Threshold
pixel$noiseRm                       <- ifelse(pixel$radiance<0, NA, pixel$radiance)

## Step 3 - Remove outliers
pixel$noiseRm_out                   <- pixel$noiseRm %>% tsclean(replace.missing = FALSE)    

## Step 5 - Bias correction
pixel$noiseRm_out_bc                <- bias_correction(pixel$noiseRm_out, pixel$n.images) %>% as.vector()

## Step 6 - Interpolation
pixel$noiseRm_out_bc_interpolated   <- na.interp(pixel$noiseRm_out_bc)

## Conventional cleaning i.e. outlier removal, threshold and interpolation
pixel$noiseRm_out_interpolated      <- na.interp(pixel$noiseRm_out)



## Generating plots according to the paper.

## Figure 2 - Income juxtaposed with raw nighttime lights, an example: Bombay

radiance_income_plot(radiance=rawAgg, income = incomeAgg, 
                            pngFilename = "../RESULTS/radiance_rawPixelLevel_incomeReal_HR_Bombay.png",
                             legendTitle=c(rawLeg, incLeg), color=c(col_rawRad, col_income), lty_type=c(lty_rawRad, lty_income))


## Figure 4 Seasonal dips in nighttime lights, an example: one pixel in Bombay
radiance_cfobs_plot(radiance=pixel$noiseRm_out_interpolated, cfobs=pixel$n.images,
                pngFilename="../RESULTS/radiance_cfobs_px1_noiseRm_out_interpolated.png", vline=TRUE,
                color=c(col_conventionalRad, "black"), lty_type=c(lty_conventionalRad, 1))



## Figure 5 Seasonal dips in nighttime lights, an example: Bombay
# Radiance and cloud-free observations plot - Bombay (HR)
radiance_cfobs_plot(radiance=conventionalAgg, cfobs=cloudsAgg,
                    pngFilename="../RESULTS/radiance_cfobs_noiseRm_out_interpolated_HR_Bombay.png", vline=TRUE,
                    color=c(col_conventionalRad, "black"), lty_type=c(lty_conventionalRad, 1))



## Figure 6 Economic significance of seasonal patterns, an example: Bombay
monthlySeasonality_Bombay       <- monthlySeasonality_df(rawAgg, conventionalAgg, 
                                biasCorrectedAgg, worldbankAgg , incomeAgg)


monthlySeas_lineplot(df=monthlySeasonality_Bombay, legendTitle=c(incLeg, conLeg),
        color=c(col_income, col_conventionalRad), lty_type=c(lty_income, lty_conventionalRad),
        pngFilename="../RESULTS/seasonality_incomeReal_conventional_HR_Bombay_lineplot.png")


## Figure 8 Nighttime lights vs cloud-free observation, an example: one pixel in Bombay
        
## Generating the loess plot
date <- seq(as.yearmon("2014-01"), by=1/12, length.out=length(pixel$radiance))
time <- date - as.yearmon("2014-01")

# Step 1: Remove the time trend
ly1             <- log(pixel$noiseRm_out)              # ly1 is the conventionally cleaned series
m.timetrend     <- lm(ly1 ~ time, na.action=na.exclude)
ly2             <- residuals(m.timetrend)               # ly2 is the residuals after removing the time trend
cf              <- pixel$n.images
# Applying loess function
m.loess         <- loess(ly2 ~ cf, na.action=na.exclude)
tmp = data.frame(cf, predict(m.loess))
tmp = tmp[order(tmp$predict.m.loess),]
## Plotting loess
png("../RESULTS/radiance_cfobs_px1_noiseRm_out_interpolated_timeRem_loess.png", width=800)
par(mfrow=c(1,1))
par(mai=c(1.0,1.2,.4,.4))
par(mfrow=c(1,1))
par(family = TexFont)
yrange  <- range(c(min(na.omit(ly2)), max(na.omit(ly2))), na.rm=TRUE)
plot(cf, ly2, type="p", lwd=3, ylim = yrange, 
    ylab=radUnitsLog, 
    xlab="Cloud-free observations", main = "", cex.lab=cexlab, cex.main=2, cex.axis=cexaxis)
points(arrange(data.frame(cf, predict(m.loess)), cf), type="l", col=col_conventionalRad, lwd=3)
legend(x="topright", bty="n", lwd=3, col=c(col_conventionalRad), legend=TeX("Loess") ,cex=cexlegend*0.8)
abline(h=0)
grid(lwd=3)
dev.off()

## Figure 9 An example of bias-corrected nighttime lights: Bombay

radiance_plot_2series(radiance=conventionalAgg, radiance2=biasCorrectedAgg,
                            pngFilename = "../RESULTS/radiance_conventional_bcPixelLevel_HR_Bombay.png",
                            title = "", legendTitle=c(conLeg, bcLeg), ymin=30000, ymax=75000, color=c(col_conventionalRad, col_bcRad), lty_type=c(lty_conventionalRad, lty_bcRad))


## Figure 10 Economic significance of seasonal patterns, an example: Bombay

monthlySeas_lineplot_4series(df=monthlySeasonality_Bombay, legendTitle=c(incLeg, bcLeg, "World Bank method", conLeg),
lty_type=c(lty_income, lty_bcRad, lty_worldbankRad, lty_conventionalRad), color=c(col_income, col_bcRad, col_worldbankRad, col_conventionalRad), 
        pngFilename="../RESULTS/seasonality_incomeReal_conventional_bc_worldbank_HR_Bombay_lineplot.png")


